//
// GameX - WindowsDX Class Code (including WinMain and WinProc)
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "gamex-win-dx8.hpp"

// WindowsDX Constructor
WindowsDX::WindowsDX (int nWinFull, int xres, int yres)
{
	windx_full = nWinFull;
	windx_xsize = xres;
	windx_ysize = yres;
	table_alpha = NULL;
}

WindowsDX::~WindowsDX (void)
{
	if (table_alpha!=NULL) delete [] table_alpha;
}

// WinMain and WindowFunc functions supplied by WindowsDX

LRESULT CALLBACK WindowFunc (HWND hWnd,	UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_ACTIVATEAPP: GameX.SetActivation ((BOOL) wParam); break;
	case WM_MOVE: GameX.SetPosition (LOWORD(lParam), HIWORD(lParam)); break;
	case WM_CHAR: {
		if (((TCHAR )wParam) == 27) PostQuitMessage (0);
		GameX.SetKey ((TCHAR) wParam); 
	} break;
	case WM_KEYDOWN: GameX.SetKeyDown ((TCHAR) wParam); break;		
	case WM_KEYUP: GameX.SetKeyUp ((TCHAR) wParam); break;
	case WM_CREATE:	break;
	case WM_SETFOCUS: {
		if (GameX.DrawOk()) {
			GameX.SetActivation (true);
			GameX.ViewUpdate(); 		
		}
	} break;
	case WM_KILLFOCUS: GameX.SetActivation(false); break;
	case WM_PAINT: break;
	case WM_SIZE: break;
	case WM_CLOSE: DestroyWindow (hWnd); break;
	case WM_DESTROY: PostQuitMessage (0); break;
	case WM_MOUSEMOVE: {
		GameX.SetMouseMove (LOWORD (lParam), HIWORD (lParam));
		if (wParam != 0) {
			//if (wParam & MK_LBUTTON) {
			//} else if (wParam & MK_MBUTTON) {
			//} else if (wParam & MK_RBUTTON) {
		}						
	} break;
	}
	return DefWindowProc (hWnd, message, wParam, lParam);		
}

int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	MSG msg;
	if (GameX.InitWindow (hThisInst, nWinMode, " ")==FALSE) return FALSE;
	GameX.InitDirect ();

	initialize ();

	for (;;) {
		if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			main();
		}
	}
	GameX.CleanUp ();
	return msg.wParam;
}

// WindowsDX Member Functions


int WindowsDX::InitWindow (HINSTANCE hThisInst,
						   int nWinMode, 
						   char *szWinName)
{
	debug.Open ("debug.txt");
	debug.Output ("WindowsDX: Initialization started.\n", "");
	win_drawok = 0;
	if (WindowInitClass (hThisInst, szWinName)==FALSE)		return FALSE;
	if (WindowInit (hThisInst, szWinName)==FALSE) return FALSE;
	if (WindowDraw (nWinMode)==FALSE)						return FALSE;

	return TRUE;
}

int WindowsDX::InitDirect (void)
{	
	if (DirectDrawInit ()==FALSE)							return FALSE;
	if (DirectDrawSurfInit ()==FALSE)						return FALSE;
	debug.Output ("WindowsDX: Initialization finished ok.\n", "");

	win_keys = 0;
	win_drawok = 1;

	return TRUE;
}
		
int WindowsDX::WindowInitClass (HINSTANCE hThisInst,		
								char *szWinName)				
{
	WNDCLASS wcl;

  	wcl.hCursor			= LoadCursor(NULL,IDC_ARROW);
	wcl.hIcon			= NULL;
    wcl.lpszMenuName	= "Menu";
    wcl.lpszClassName	= szWinName;
	wcl.hInstance		= hThisInst;
	wcl.style			= CS_VREDRAW | CS_HREDRAW;
    wcl.lpfnWndProc		= (WNDPROC) WindowFunc;
	wcl.cbClsExtra		= 0;
    wcl.cbWndExtra		= 0;
    wcl.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
    
    if (!RegisterClass(&wcl)) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::WindowInitClass: Unable to register Windows class.\n", "");
		#endif	
		return FALSE;		
	}

	return TRUE;
}

int WindowsDX::WindowInit (HINSTANCE hThisInst, char *szWinName)
{
	DWORD style;
	HDC hdc;

	// Initialize View Format (Bits-Per-Pixel)
	hdc = GetDC (win_hwnd);
	win_screenbpp = GetDeviceCaps (hdc, BITSPIXEL);
	ReleaseDC (win_hwnd, hdc);
	
	// Initialize View Size & FullScreen Mode
	win_screenx = GetSystemMetrics(SM_CXSCREEN);
	win_screeny = GetSystemMetrics(SM_CYSCREEN);
	win_fullscreen = windx_full;

	// Initialize View Width & Height
	if (win_fullscreen==TRUE)	{
		style = WS_POPUP;	
		win_width = windx_xsize;
        win_height = windx_ysize;
	} else {
		style = WS_OVERLAPPED | WS_SYSMENU;
		win_width = windx_xsize;		
		win_height = windx_ysize;		
		if (win_width<=0 || win_width>win_screenx) win_width = win_screenx;
		if (win_height<=0 || win_height>win_screeny) win_height = win_screeny;			
	}

	// Initialize Drawing Viewport
	win_viewx1 = 0;
	win_viewy1 = 0;
	win_viewx2 = win_width-1;
	win_viewy2 = win_height-1;

	// Initialize Drawing Dispact Functions
	switch (win_screenbpp) {
	case 16: {
		FuncDrawPixel = &DrawPixel16;
		FuncDrawImage = &DrawImage16;
		FuncDrawMasked = &DrawMasked16;
		FuncDrawAlpha = &DrawAlpha16;
		FuncDrawAdded = &DrawAdded16;
	} break;
	default: {
		FuncDrawPixel = NULL;
		FuncDrawImage = NULL;
		FuncDrawMasked = NULL;
		FuncDrawAlpha = NULL;
		FuncDrawAdded = NULL;
	} break;
	}
		
	// Create window adjusting for desired client area size
	// NOTE: For some reason (windows crap??) the AdjustWindowRectEx function
	// is not working.

	RECT	rect;
	int		sx, sy;
	HWND	temp_hwnd;

	// Create temporary window to determine extended window size
	temp_hwnd = CreateWindowEx(WS_EX_APPWINDOW, szWinName, szWinName, style, 0, 0, win_width, win_height, 0, 0, hThisInst, 0);
	if (temp_hwnd==NULL) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::WindowInit: Unable to create temp window.\n", "");
		#endif		
		return FALSE;
	}
	GetClientRect (temp_hwnd, &rect);
	
	sx = (win_width<<1) - rect.right;
	sy = (win_height<<1) - rect.bottom;
	if (win_fullscreen==TRUE) {
		rect.left = 0; rect.top = 0;
	} else {
		rect.left = (win_screenx>>1) - (sx>>1);
		rect.top = (win_screeny>>1) - (sy>>1);
	}	
	rect.right = rect.left + sx-1;		
	rect.bottom = rect.top + sy-1;

	// Create actual window (with correct size) which will be used

	win_hwnd = CreateWindowEx(
					WS_EX_APPWINDOW, 
					szWinName, 
					szWinName, 
					style, 
					rect.left, rect.top, 
					rect.right-rect.left, rect.bottom-rect.top,
					0, 
					0, 
					hThisInst, 
					0);
	if (win_hwnd==NULL) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::WindowInit: Unable to create window.\n", "");
		#endif		
		return FALSE;
	}

	// Double check to make sure we have correct client rect size	
	GetClientRect (win_hwnd, &rect);	
	
	return TRUE;
}

int WindowsDX::WindowDraw (int nWinMode)
{
	ShowWindow (win_hwnd, nWinMode);
	if (UpdateWindow (win_hwnd)==FALSE) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::WindowDraw: Unable to update Window.\n", "");
		#endif		
		return FALSE;		
	}	
	return TRUE;
}

int WindowsDX::DirectDrawInit ()
{
	HRESULT status;
	int style;

	// Direct Draw 8.0
	/* dd_main = NULL; 
	dd_device = NULL; 
	dd_sprite = NULL;
	dd_caps = new D3DCAPS8;

	// Create the D3D object, which is needed to create the D3DDevice.
	dd_main = Direct3DCreate8(D3D_SDK_VERSION);

	// Get the current desktop display mode
    D3DDISPLAYMODE d3ddm;
	
	dd_main->GetAdapterDisplayMode (D3DADAPTER_DEFAULT, &d3ddm);

	// Parameters for the D3DDevice. Most parameters are zero'd out.
    // This sets the video format to match the current desktop display.
    // Check docs for other options you can set here, such as
    // 'Windowed' which would create a window-based app (not full screen)
    
	D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory (&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.hDeviceWindow = win_hwnd;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferWidth = win_screenx;		//d3ddm.Width;
    d3dpp.BackBufferHeight = win_screeny;		//d3ddm.Height;
    d3dpp.BackBufferFormat = d3ddm.Format;

	// Create the Direct3D device, using the default adapter (most systems 
    // only have one, unless they have multiple graphics hardware cards
    // installed). See SDK for more details on other options here.
    
	dd_main->CreateDevice (D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, win_hwnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &dd_device);
	
	dd_device->SetRenderState(D3DRS_LIGHTING, FALSE); */

	// Direct Draw 3.0	
	status = DirectDrawCreate (NULL, &dd_main, NULL);
	if (status!=DD_OK) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DirectDrawInit: Unable to initialize Direct Draw object.\n", "");
		#endif		
		return FALSE;
	}

	if (win_fullscreen==TRUE) {
		style = DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE;
	} else {
		style = DDSCL_NORMAL;
	}

	status = dd_main->SetCooperativeLevel (win_hwnd, style);
	if (status!=DD_OK) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DirectDrawInit: Unable to set cooperative level.\n", "");
		#endif		
		return FALSE;
	}

	return TRUE; 
}

int WindowsDX::DirectDrawSurfInit ()
{	
    // Direct X 8.1
	/*D3DCOLOR			colorkey = 0xFF000000;
    D3DXIMAGE_INFO		info; 
	HRESULT				result;
	D3DDISPLAYMODE d3ddm;
	
	dd_main->GetAdapterDisplayMode (D3DADAPTER_DEFAULT, &d3ddm);

	D3DXCreateSprite(dd_device, &dd_sprite);

	result = D3DXCreateTextureFromFileEx (dd_device, "horse.jpg", 0, 0, 1, 0,
		d3ddm.Format, D3DPOOL_MANAGED, D3DX_FILTER_NONE, D3DX_DEFAULT, 
        colorkey, &info , NULL, &dd_texture);
	if (result!=D3D_OK) exit (-22);
 
	dd_rect = new RECT;
	dd_rect->left = 0;
	dd_rect->top  = 0;
	dd_rect->bottom = 256;
	dd_rect->right = 256; */

	// DirectX 3.1
	HRESULT status;
	DDSURFACEDESC ddsd;
	POINT pt;

	if (dd_back_buf)	dd_back_buf->Release (),	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release (),	dd_front_buf = NULL;	

	ZeroMemory (&ddsd, sizeof (ddsd));
	
	if (win_fullscreen==TRUE) {
		// Set display mode
	 	status = dd_main->SetDisplayMode (win_width, win_height, win_screenbpp);
		if (status!=DD_OK) return FALSE;
		// Set front surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.dwBackBufferCount = 1;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {
			switch (status) {
			case DDERR_INVALIDOBJECT:			debug.Output ("<WinDX::CreateSurface> Invalid Object\n","");break;
			case DDERR_INVALIDPARAMS:			debug.Output ("<WinDX::CreateSurface> Invalid Params\n","");break;
			case DDERR_OUTOFVIDEOMEMORY:		debug.Output ("<WinDX::CreateSurface> Out Of Video Memory\n","");break;
			case DDERR_NODIRECTDRAWHW:			debug.Output ("<WinDX::CreateSurface> No Direct Draw HW\n","");break;
			case DDERR_NOCOOPERATIVELEVELSET:	debug.Output ("<WinDX::CreateSurface> No Cooperate Level Set\n","");break;
			case DDERR_INVALIDCAPS:				debug.Output ("<WinDX::CreateSurface> Invalid Caps\n","");break;
			case DDERR_INVALIDPIXELFORMAT:		debug.Output ("<WinDX::CreateSurface> Invalid Pixel Format\n","");break;
			case DDERR_NOALPHAHW:				debug.Output ("<WinDX::CreateSurface> No Alpha HW\n","");break;
			case DDERR_NOFLIPHW:				debug.Output ("<WinDX::CreateSurface> No Flip HW\n","");break;
			case DDERR_NOZBUFFERHW:				debug.Output ("<WinDX::CreateSurface> No Z-Buffer HW\n","");break;
			case DDERR_NOEXCLUSIVEMODE:			debug.Output ("<WinDX::CreateSurface> No Exclusive Mode\n","");break;
			case DDERR_OUTOFMEMORY: 			debug.Output ("<WinDX::CreateSurface> Out of Memory\n","");break;
			case DDERR_PRIMARYSURFACEALREADYEXISTS:	debug.Output ("<WinDX::CreateSurface> Primary Surface Already Exists\n","");break;
			case DDERR_NOEMULATION:				debug.Output ("<WinDX::CreateSurface> No Emulation\n","");		   break;
			case DDERR_INCOMPATIBLEPRIMARY:		debug.Output ("<WinDX::CreateSurface> Incompatible Primary\n",""); break;
			}
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);		
		}			
		// No good, return error
		if (status!=DD_OK) {
			#ifdef WINDX_DEBUG
				debug.Output ("WinDX::DirectDrawSurfInit: Unable to create primary surface.\n", "");
			#endif		
			return FALSE;
		}
		// Set back surface description
	   	DDSCAPS caps;		
		caps.dwCaps = DDSCAPS_BACKBUFFER;
		status = dd_front_buf->GetAttachedSurface (&caps, &dd_back_buf);
		if (status!=DD_OK) return FALSE;
	} else {		
		// Set front surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_VIDEOMEMORY;
		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {			
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		}
		// No good, return error
		if (status!=DD_OK) {
			#ifdef WINDX_DEBUG
				debug.Output ("WinDX::DirectDrawSurfInit: Unable to create primary surface.\n", "");
			#endif		
			return FALSE;
		}	
		// Set back surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth = win_width;
		ddsd.dwHeight = win_height;
		status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
		}
		// No good, return error
		if (status!=DD_OK) {
			#ifdef WINDX_DEBUG
				debug.Output ("WinDX::DirectDrawSurfInit: Unable to create secondary surface.\n", "");
			#endif		
			return FALSE;
		}
	} 

    // Calculate surface (buffer) sizes, widths and pitch
	ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_PITCH;
    dd_back_buf->GetSurfaceDesc(&ddsd);
	win_pitch = ddsd.lPitch;
	if (win_screenbpp==16)	win_width = win_pitch>>1;
	else					win_width = ddsd.dwWidth;
	win_height = ddsd.dwHeight;	
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {		
		debug.Output ("WinDX:: Back Buffer is in video memory.\n", "");
	} else {
		debug.Output ("WinDX:: Back Buffer NOT IN VIDEO MEMORY.\n", "");
	}
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_BACKBUFFER) {
		debug.Output ("WinDX:: Back Buffer is a valid flip (back) buffer.\n", "");
	} else {
		debug.Output ("WinDX:: Back Buffer NOT A FLIP (BACK) BUFFER.\n", "");
	}
	dd_front_buf->GetSurfaceDesc(&ddsd);
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {
		debug.Output ("WinDX:: Front Buffer is in video memory.\n", "");
	} else {
		debug.Output ("WinDX:: Front Buffer NOT IN VIDEO MEMORY.\n", "");
	}	

	// Precompute: y-table look up 
	for (int y=0; y<=win_height; y++)
		win_ytable[y] = y*win_pitch;
		
	// Precompute: screen size (total number of bytes)
	win_size = win_width*win_height;
	if (win_screenbpp==16)	win_size <<= 1;	
	
	// Precompute: client rectangle
	
	GetClientRect(win_hwnd, &win_destrect);
    pt.x = pt.y = 0;
    ClientToScreen(win_hwnd, &pt);
	OffsetRect(&win_destrect, pt.x, pt.y);
	win_xpos = pt.x; win_ypos = pt.y;

	win_active = TRUE;
	

    return TRUE;
}

int WindowsDX::DirectDrawCleanUp (void)
{
	// Direct Draw 8.1
	/*
	if (dd_device) dd_device->Release();
    if (dd_main) dd_main->Release();	
	if (dd_caps) delete dd_caps; */

	// Direct Draw 3.0
	if (dd_back_buf)	dd_back_buf->Release (),	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release (),	dd_front_buf = NULL;	
	if (dd_main)		dd_main->Release (),		dd_main = NULL;
	return TRUE;
}

int WindowsDX::CleanUp (void)
{
	if (DirectDrawCleanUp()==FALSE)	{
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::CleanUp: Unable to close Direct Draw.\n", "");
		#endif			
		return FALSE;
	}
	return TRUE;
}

int WindowsDX::ViewUpdate (void)
{
    // Direct X 8.1
	//dd_device->Present (NULL, NULL, win_hwnd, NULL);

	// Direct Draw 3.0
	HRESULT             status;	
	DDBLTFX				bltfx;	
	
	bltfx.dwSize = sizeof (DDBLTFX);
	bltfx.dwDDFX = DDBLTFX_NOTEARING;
	
	
    if (win_active==TRUE) {
		if (GetForegroundWindow() == win_hwnd) {
			if (win_fullscreen==TRUE) {
				dd_front_buf->Flip(NULL, DDFLIP_WAIT);	// Flip back buffer to front
				if (dd_front_buf->IsLost() == DDERR_SURFACELOST) 
					dd_front_buf->Restore();
			} else {				
				status = dd_front_buf->BltFast (win_xpos, win_ypos, dd_back_buf, 0, DDBLTFAST_WAIT);
				if (status==DD_OK) return TRUE;
				if (status==DDERR_WASSTILLDRAWING) return FALSE;
				if (status==DDERR_SURFACELOST) {
					MessageBox (win_hwnd, "<WinDX::ViewUpdate> Error: Surface Lost", "",MB_OK);
					status = dd_front_buf->Restore ();
					if (status!=DD_OK) return FALSE;
					status = dd_back_buf->Restore ();
					if (status!=DD_OK) return FALSE;				
				}				
			}			
		}		
	}
	return TRUE;
}


inline void WindowsDX::SetActivation (BOOL act)
{
	win_active = act;
}

inline void WindowsDX::SetPosition (int x, int y)
{
	win_xpos = x;
	win_ypos = y;
}

inline int WindowsDX::DrawOk (void)
{
	return win_drawok;
}

void WindowsDX::SetView (int x1, int y1, int x2, int y2)
{
	win_viewx1 = x1;
	win_viewy1 = y1;
	win_viewx2 = x2;
	win_viewy2 = y2;
}

inline void WindowsDX::SetKey (char ch)
{
	win_keybuf[win_keys++] = ch;
	if (win_keys > WINDX_BUFSIZE) {
		win_keys = WINDX_BUFSIZE;
		debug.Output ("WinDX: Keyboard buffer overflow.\n", "");
	}
}

char WindowsDX::GetKey (void) 
{	
	if (win_keys>0) {
		win_keys--;		
		return win_keybuf[win_keys];
    } else {
		return 0;
	}	
}

int WindowsDX::DrawOpen (void)
{
	/*HRESULT status;

	status = dd_device->BeginScene();
	if (status!=D3D_OK) exit (-21); */

	// Direct Draw 3.0
	HRESULT				status;
	DDSURFACEDESC		ddsd;

	ddsd.dwSize = sizeof(ddsd);
	status = dd_back_buf->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);	// lock drawing surface
	if (status!=DD_OK) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawOpen: Unable to lock direct draw surface.\n", "");
		#endif	
		return FALSE;
	}

	win_screen = (char *) ddsd.lpSurface;
	return TRUE;
}

int WindowsDX::DrawClose (void)
{
	// Direct X 8.1
	//dd_device->EndScene();

	// Direct Draw 3.0
	dd_back_buf->Unlock (NULL);							// unlock drawing surface
	return TRUE;
}

void WindowsDX::DirectDrawError (int status)
{
	switch (status) {
	case DDERR_ALREADYINITIALIZED: debug.Output ("DirectX: This object is already initialized.\n",""); break;
	case DDERR_BLTFASTCANTCLIP: debug.Output ("DirectX: Return if a clipper object is attached to the source surface passed into a BltFast call.\n",""); break;
	case DDERR_CANNOTATTACHSURFACE: debug.Output ("DirectX: This surface can not be attached to the requested surface.\n",""); break; 
	case DDERR_CANNOTDETACHSURFACE: debug.Output ("DirectX: This surface can not be detached from the requested surface.\n","");break;
	case DDERR_CANTCREATEDC: debug.Output ("DirectX: Windows can not create any more DCs\n",""); break;
	case DDERR_CANTDUPLICATE: debug.Output ("DirectX: Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.\n",""); break;
	case DDERR_CLIPPERISUSINGHWND: debug.Output ("DirectX: An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd.\n",""); break;
	case DDERR_COLORKEYNOTSET: debug.Output ("DirectX: No src color key specified for this operation.\n",""); break;
	case DDERR_CURRENTLYNOTAVAIL: debug.Output ("DirectX: Support is currently not available.\n",""); break;
	case DDERR_DIRECTDRAWALREADYCREATED: debug.Output ("DirectX: A DirectDraw object representing this driver has already been created for this process.\n",""); break;
	case DDERR_EXCEPTION: debug.Output ("DirectX: An exception was encountered while performing the requested operation.\n",""); break;
	case DDERR_EXCLUSIVEMODEALREADYSET: debug.Output ("DirectX: An attempt was made to set the cooperative level when it was already set to exclusive.\n",""); break;
	case DDERR_GENERIC: debug.Output ("DirectX: Generic failure.\n",""); break;
	case DDERR_HEIGHTALIGN: debug.Output ("DirectX: Height of rectangle provided is not a multiple of reqd alignment.\n",""); break;
	case DDERR_HWNDALREADYSET: debug.Output ("DirectX: The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.\n",""); break;
	case DDERR_HWNDSUBCLASSED: debug.Output ("DirectX: HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.\n",""); break;
	case DDERR_IMPLICITLYCREATED: debug.Output ("DirectX: This surface can not be restored because it is an implicitly created surface.\n",""); break;
	case DDERR_INCOMPATIBLEPRIMARY: debug.Output ("DirectX: Unable to match primary surface creation request with existing primary surface.\n",""); break;
	case DDERR_INVALIDCAPS: debug.Output ("DirectX: One or more of the caps bits passed to the callback are incorrect.\n",""); break;
	case DDERR_INVALIDCLIPLIST: debug.Output ("DirectX: DirectDraw does not support the provided cliplist.\n",""); break;
	case DDERR_INVALIDDIRECTDRAWGUID: debug.Output ("DirectX: The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.\n",""); break;
	case DDERR_INVALIDMODE: debug.Output ("DirectX: DirectDraw does not support the requested mode.\n",""); break;
	case DDERR_INVALIDOBJECT: debug.Output ("DirectX: DirectDraw received a pointer that was an invalid DIRECTDRAW object.\n",""); break;
	case DDERR_INVALIDPARAMS: debug.Output ("DirectX: One or more of the parameters passed to the function are incorrect.\n",""); break;
	case DDERR_INVALIDPIXELFORMAT: debug.Output ("DirectX: The pixel format was invalid as specified.\n",""); break;
	case DDERR_INVALIDPOSITION: debug.Output ("DirectX: Returned when the position of the overlay on the destination is no longer legal for that destination.\n",""); break;
	case DDERR_INVALIDRECT: debug.Output ("DirectX: Rectangle provided was invalid.\n",""); break;
	case DDERR_LOCKEDSURFACES: debug.Output ("DirectX: Operation could not be carried out because one or more surfaces are locked.\n",""); break;
	case DDERR_NO3D: debug.Output ("DirectX: There is no 3D present.\n",""); break;
	case DDERR_NOALPHAHW: debug.Output ("DirectX: Operation could not be carried out because there is no alpha accleration hardware present or available.\n",""); break;
	//case DDERR_NOANTITEARHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for synchronizing blits to avoid tearing.	\n",""); break;
	case DDERR_NOBLTHW: debug.Output ("DirectX: No blitter hardware present.\n",""); break;
	//case DDERR_NOBLTQUEUEHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for asynchronous blitting.\n",""); break;
	case DDERR_NOCLIPLIST: debug.Output ("DirectX: No cliplist available.\n",""); break;
	case DDERR_NOCLIPPERATTACHED: debug.Output ("DirectX: No clipper object attached to surface object.\n",""); break;
	case DDERR_NOCOLORCONVHW: debug.Output ("DirectX: Operation could not be carried out because there is no color conversion hardware present or available.\n",""); break;
	case DDERR_NOCOLORKEY: debug.Output ("DirectX: Surface doesn't currently have a color key\n",""); break;
	case DDERR_NOCOLORKEYHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support of the destination color key.\n",""); break;
	case DDERR_NOCOOPERATIVELEVELSET: debug.Output ("DirectX: Create function called without DirectDraw object method SetCooperativeLevel being called.\n",""); break;
	case DDERR_NODC: debug.Output ("DirectX: No DC was ever created for this surface.\n",""); break;
	case DDERR_NODDROPSHW: debug.Output ("DirectX: No DirectDraw ROP hardware.\n",""); break;
	case DDERR_NODIRECTDRAWHW: debug.Output ("DirectX: A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware.\n",""); break;
	case DDERR_NOEMULATION: debug.Output ("DirectX: Software emulation not available.\n",""); break;
	case DDERR_NOEXCLUSIVEMODE: debug.Output ("DirectX: Operation requires the application to have exclusive mode but the application does not have exclusive mode.\n",""); break;
	case DDERR_NOFLIPHW: debug.Output ("DirectX: Flipping visible surfaces is not supported.\n",""); break;
	case DDERR_NOGDI: debug.Output ("DirectX: There is no GDI present.\n",""); break;
	case DDERR_NOHWND: debug.Output ("DirectX: Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.\n",""); break;
	case DDERR_NOMIRRORHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware present or available.\n",""); break;
	case DDERR_NOOVERLAYDEST: debug.Output ("DirectX: Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination.\n",""); break;
	case DDERR_NOOVERLAYHW: debug.Output ("DirectX: Operation could not be carried out because there is no overlay hardware present or available.\n",""); break;
	case DDERR_NOPALETTEATTACHED: debug.Output ("DirectX: No palette object attached to this surface.\n",""); break;
	case DDERR_NOPALETTEHW: debug.Output ("DirectX: No hardware support for 16 or 256 color palettes.\n",""); break;
	case DDERR_NORASTEROPHW: debug.Output ("DirectX: Operation could not be carried out because there is no appropriate raster op hardware present or available.\n",""); break;
	case DDERR_NOROTATIONHW: debug.Output ("DirectX: Operation could not be carried out because there is no rotation hardware present or available.\n",""); break;
	case DDERR_NOSTRETCHHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for stretching.\n",""); break;
	case DDERR_NOT4BITCOLOR: debug.Output ("DirectX: DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.\n",""); break;
	case DDERR_NOT4BITCOLORINDEX: debug.Output ("DirectX: DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.\n",""); break;
	case DDERR_NOT8BITCOLOR: debug.Output ("DirectX: DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color.\n",""); break;
	case DDERR_NOTAOVERLAYSURFACE: debug.Output ("DirctX: Returned when an overlay member is called for a non-overlay surface.\n",""); break;
	case DDERR_NOTEXTUREHW: debug.Output ("DirectX: Operation could not be carried out because there is no texture mapping hardware present or available.\n",""); break;
	case DDERR_NOTFLIPPABLE: debug.Output ("DirectX: An attempt has been made to flip a surface that is not flippable.\n",""); break;
	case DDERR_NOTFOUND: debug.Output ("DirectX: Requested item was not found.\n",""); break;
	case DDERR_NOTLOCKED: debug.Output ("DirectX: Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.\n",""); break;
	case DDERR_NOTPALETTIZED: debug.Output ("DirectX: The surface being used is not a palette-based surface.\n",""); break;
	case DDERR_NOVSYNCHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.\n",""); break;
	case DDERR_NOZBUFFERHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for zbuffer blitting.\n",""); break;
	case DDERR_NOZOVERLAYHW: debug.Output ("DirectX: Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.\n",""); break;
	case DDERR_OUTOFCAPS: debug.Output ("DirectX: The hardware needed for the requested operation has already been allocated.\n",""); break;
	case DDERR_OUTOFMEMORY: debug.Output ("DirectX: DirectDraw does not have enough memory to perform the operation.\n",""); break;
	case DDERR_OUTOFVIDEOMEMORY: debug.Output ("DirectX: DirectDraw does not have enough memory to perform the operation.\n",""); break;
	case DDERR_OVERLAYCANTCLIP: debug.Output ("DirectX: The hardware does not support clipped overlays.\n",""); break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: debug.Output ("DirectX: Can only have ony color key active at one time for overlays.\n",""); break;
	case DDERR_OVERLAYNOTVISIBLE: debug.Output ("DirectX: Returned when GetOverlayPosition is called on a hidden overlay.\n",""); break;
	case DDERR_PALETTEBUSY: debug.Output ("DirectX: Access to this palette is being refused because the palette is already locked by another thread.\n",""); break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS: debug.Output ("DirectX: This process already has created a primary surface.\n",""); break;
	case DDERR_REGIONTOOSMALL: debug.Output ("DirectX: Region passed to Clipper::GetClipList is too small.\n",""); break;
	case DDERR_SURFACEALREADYATTACHED: debug.Output ("DirectX: This surface is already attached to the surface it is being attached to.\n",""); break;
	case DDERR_SURFACEALREADYDEPENDENT: debug.Output ("DirectX: This surface is already a dependency of the surface it is being made a dependency of.\n",""); break;
	case DDERR_SURFACEBUSY: debug.Output ("DirectX: Access to this surface is being refused because the surface is already locked by another thread.\n",""); break;
	case DDERR_SURFACEISOBSCURED: debug.Output ("DirectX: Access to surface refused because the surface is obscured.\n",""); break;
	case DDERR_SURFACELOST: debug.Output ("DirectX: Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it.\n",""); break;
	case DDERR_SURFACENOTATTACHED: debug.Output ("DirectX: The requested surface is not attached.\n",""); break;
	case DDERR_TOOBIGHEIGHT: debug.Output ("DirectX: Height requested by DirectDraw is too large.\n",""); break;
	case DDERR_TOOBIGSIZE: debug.Output ("DirectX: Size requested by DirectDraw is too large --	the individual height and width are OK.\n",""); break;
	case DDERR_TOOBIGWIDTH: debug.Output ("DirectX: Width requested by DirectDraw is too large.\n",""); break;
	case DDERR_UNSUPPORTED: debug.Output ("DirectX: Action not supported.\n",""); break;
	case DDERR_UNSUPPORTEDFORMAT: debug.Output ("DirectX: FOURCC format requested is unsupported by DirectDraw.\n",""); break;
	case DDERR_UNSUPPORTEDMASK: debug.Output ("DirectX: Bitmask in the pixel format requested is unsupported by DirectDraw.\n",""); break;
	case DDERR_VERTICALBLANKINPROGRESS: debug.Output ("DirectX: Vertical blank is in progress.\n",""); break;
	case DDERR_WASSTILLDRAWING: debug.Output ("DirectX: Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.\n",""); break;
	case DDERR_WRONGMODE: debug.Output ("DirectX: This surface can not be restored because it was created in a different mode.\n",""); break;
	}
}
void WindowsDX::DrawClear (int ynum)
{	
	unsigned short *start;
	int xres;
	start = (unsigned short *) win_screen;
	xres = win_pitch;
	for (int y=0; y<ynum; y++) {			
		memset (start, 10, xres);
		start += win_width;		
	}
}



void WindowsDX::DrawColor (int r, int g, int b)
{
	win_r = r;	win_g = g;	win_b = b;
	win_rgb = RGB(win_r, win_g, win_b);
}

bool WindowsDX::ClipImage (int x, int y, ImageX *img, int &view_offset, int &img_offset, int &img_xs, int &img_ys)
{
	int x2, y2;

	img_xs = img->m_xres;
	img_ys = img->m_yres;
	x2 = x+(img_xs-1);
	y2 = y+(img_ys-1);
	
	if (x2 >= win_viewx1 && x <= win_viewx2) {			// In-View Horizontally
		if (y2 >= win_viewy1 && y <= win_viewy2) {		// In-View Vertically 
			view_offset = win_viewy1*win_width + win_viewx1;
			img_offset = 0;			
			if (x < win_viewx1) {						// Clip Left
				img_xs -= (win_viewx1 - x);
				img_offset += (win_viewx1 - x);
			} else {
				view_offset += (x - win_viewx1);
			}
			if (x2 > win_viewx2)						// Clip Right
				img_xs -= (x2 - win_viewx2);
			if (y < win_viewy1) {						// Clip Top
				img_ys -= (win_viewy1 - y);
				img_offset += (win_viewy1 - y) * img->m_xres;	
			} else {
				view_offset += (y - win_viewy1) * win_width;
			}
			if (y2 > win_viewy2)						// Clip Bottom
				img_ys -= (y2 - win_viewy2);
			return true;
		}
	}
	return false;			
}

void WindowsDX::UseAlpha (void)
{
	int a, b, c;	
	table_alpha = new unsigned short[32*32];
	
	for (a=0; a<32; a++) {
		for (b=0; b<32; b++) {
			c = a+b; if (c>31) c = 31;
			table_alpha[(a<<5) + b] = c;			
			//table_alpha[(a<<5 + b)<<5] = c<<5;
			//table_alpha[a<<5 + b] = c;
		}
	}
}







void WindowsDX::DrawBox (int x1, int y1, int x2, int y2, int r, int g, int b)
{ 
	unsigned short *start, *pixel;
	int i, j, clr;

	clr = ((r>>3) << 10) + ((g>>3) << 5) + (b>>3);
	start = (unsigned short *) win_screen;
	start += y1*win_pitch + x1;
	for (j=y1; j<=y2; j++) {
		pixel = start;
		for (i=x1; i<=x2; i++)
			*pixel++ = clr;
		start += win_pitch;
	}
}

int WindowsDX::GetWidth ()
{
	return win_width;	
}

int WindowsDX::GetHeight ()
{
	return win_height;
}

int WindowsDX::GetPitch ()
{
	return win_pitch;
}

void *WindowsDX::GetScreen (void)
{
	return win_screen;
}

HWND WindowsDX::GetWindow (void)
{
	return win_hwnd;
}

int WindowsDX::DrawText (int x, int y, char *msg)
{
	HRESULT hr;
    HDC hdc = 0;
    
	hr = dd_back_buf->GetDC(&hdc);		// Get the device context handle.
    if (FAILED(hr)) { 
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawText: Unable to get device context handle.\n", "");
		#endif	
		return FALSE; 
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);
	SetTextColor(hdc, win_rgb);
    TextOut(hdc, x, y, msg, strlen(msg));
										// Release the device context.
	hr = dd_back_buf->ReleaseDC(hdc);
	return TRUE;
}

void WindowsDX::SetMouseMove (int x, int y)
{
	mouse_xl = mouse_x;
	mouse_yl = mouse_y;
	mouse_x = x;
	mouse_y = y;
	mouse_motion = TRUE;
}

void WindowsDX::SetMousePress (int button)
{
	mouse_down = TRUE;
	mouse_button = button;
}

void WindowsDX::SetMouseRelease ()
{
	mouse_down = FALSE;
	mouse_button = FALSE;
}

int WindowsDX::GetMousePress ()
{
	return mouse_button;
}

int WindowsDX::GetMouseX ()
{
	return mouse_x;
}

int WindowsDX::GetMouseY ()
{
	return mouse_y;
}

float WindowsDX::GetMouseDX ()
{
	return (float) ((mouse_x - mouse_xl)/MOUSE_POS_RATE);
}

float WindowsDX::GetMouseDY ()
{
	return (float) ((mouse_y - mouse_yl)/MOUSE_POS_RATE);
}

void WindowsDX::ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if (x<0) {
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	} else {
		mx = ((int) (x / xres));
		offx = x % xres;
	}
	if (y<0) {
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	} else {
		my = ((int) (y / yres));		
		offy = y % yres;
	}	
}

int WindowsDX::DrawBackground (int x, int y, ImageX *img)
{
	int xres, yres;
	int map_x1, map_y1;
	int map_x2, map_y2;	
	int map_numx, map_numy;	
	int off_x, off_y;
	int img_x, img_y;	
	int mx, my;

	xres = img->m_xres; yres = img->m_yres;
	ClipTile (x+win_width, y+win_height, xres, yres, map_x2, map_y2, off_x, off_y);
	ClipTile (x, y, xres, yres, map_x1, map_y1, off_x, off_y);

	map_numx = map_x2 - map_x1 + 1;
	map_numy = map_y2 - map_y1 + 1; 	
	for (my = 0; my < map_numy; my++) {
		img_y = my * yres - off_y;
		img_x = -off_x;
		for (mx = 0; mx < map_numx; mx++) {
			DrawImage (img_x, img_y, img);
			img_x += xres;
		}
	}
	return 1;
}



static int DrawPixel16 (int x, int y, int r, int g, int b)
{	
	*((unsigned short *) GameX.win_screen + GameX.win_ytable[y] + x) = (r<<10) + (g<<5) + b;
	return 1;
}

static int DrawImage16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;

	img_start = (XBYTE2 *) img->m_data;
	if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (unsigned short *) GameX.win_screen + view_offset;
		img_start += img_offset;
		for (y=0; y<img_ys; y++) {			
			memcpy (view_start, img_start, img_xs<<1);
			view_start += GameX.win_width;
			img_start += img->m_xres;
		} 		
	}
	return 1;
}

static int DrawMasked16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	unsigned short *view_start;	
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	unsigned short *view_pix;
	
	img_start = (XBYTE2 *) img->m_data;
	if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (unsigned short *) GameX.win_screen + view_offset;
		img_start += img_offset;
		for (y=0; y<img_ys; y++) {			
			img_pix = img_start;
			for (view_pix=view_start; view_pix<view_start+img_xs;) {
				if (*img_pix>0) *view_pix = *img_pix;
				view_pix++; img_pix++;
			}
			view_start += GameX.win_width;
			img_start += img->m_xres;
		} 		
	}	
	return 1;
}

static int DrawAlpha16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	XBYTE *alpha_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	XBYTE *alpha_pix;
	unsigned short *view_pix;
	register int i, v;
	
	img_start = (XBYTE2 *) img->m_data;
	alpha_start = (XBYTE *) img->m_alpha;
	if (img_start!=NULL && alpha_start!=NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (unsigned short *) GameX.win_screen + view_offset;
		img_start += img_offset;
		alpha_start += img_offset;
		for (y=0; y<img_ys; y++) {			
			img_pix = img_start;
			alpha_pix = alpha_start;
			for (view_pix=view_start; view_pix<view_start+img_xs;) {
				i = *alpha_pix;
				if (i!=255) {
					v = *view_pix;					
					*view_pix = ((v & WINDX_HIMASKB) * i) >> 8;
					*view_pix |= (((v & WINDX_HIMASKG) * i) >> 8) & WINDX_HIMASKG;
					*view_pix |= (((v >> WINDX_HISHIFTR) * i) >> 8) << WINDX_HISHIFTR;
					*view_pix += *img_pix;
				}
				view_pix++; img_pix++; alpha_pix++;
				
				/*v = *view_pix; i = *img_pix;
				c = (v & 31) + (i & 31); if (c>31) c = 31;
				*view_pix = c;
				c = (v & 992) + (i & 992); if (c>992) c = 992;
				*view_pix |= c;
				c = (v & 31774) + (i & 31774); if (c>31774) c = 31774;
				*view_pix |= c;					
				view_pix++; img_pix++;*/
			}
			view_start += GameX.win_width;
			img_start += img->m_xres;
			alpha_start += img->m_xres;
		} 		
	}
	return 1;
}

static int DrawAdded16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	XBYTE *alpha_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	XBYTE *alpha_pix;
	unsigned short *view_pix;
	register int i, v, c;
	
	img_start = (XBYTE2 *) img->m_data;
	alpha_start = (XBYTE *) img->m_alpha;
	if (img_start!=NULL && alpha_start!=NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (unsigned short *) GameX.win_screen + view_offset;
		img_start += img_offset;
		alpha_start += img_offset;
		for (y=0; y<img_ys; y++) {			
			img_pix = img_start;
			alpha_pix = alpha_start;
			for (view_pix=view_start; view_pix<view_start+img_xs;) {								
				v = *view_pix; i = *img_pix;
				if (v!=255 && i!=0) {
					c = (v & WINDX_HIMASKB) + (i & WINDX_HIMASKB); if (c>WINDX_HIMASKB) c = WINDX_HIMASKB;
					*view_pix = c;
					c = (v & WINDX_HIMASKG) + (i & WINDX_HIMASKG); if (c>WINDX_HIMASKG) c = WINDX_HIMASKG;
					*view_pix |= c;
					c = (v & WINDX_HIMASKR) + (i & WINDX_HIMASKR); if (c>WINDX_HIMASKR) c = WINDX_HIMASKR;
					*view_pix |= c;
				}
				view_pix++; img_pix++;
			}
			view_start += GameX.win_width;
			img_start += img->m_xres;
			alpha_start += img->m_xres;
		} 		
	}
	return 1;
}

